#ifndef _HUFFMAN_COMPRESSION_HPP
#define _HUFFMAN_COMPRESSION_HPP
#include "..\defines.h"
// _    _ _    _ ______ ______ __  __          _   _ 
//| |  | | |  | |  ____|  ____|  \/  |   /\   | \ | |
//| |__| | |  | | |__  | |__  | \  / |  /  \  |  \| |
//|  __  | |  | |  __| |  __| | |\/| | / /\ \ | . ` |
//| |  | | |__| | |    | |    | |  | |/ ____ \| |\  |
//|_|  |_|\____/|_|    |_|    |_|  |_/_/    \_\_| \_|
// Huffman data compression tree     
// Roger Boerdijk 30.06.2002
// email: kitt3n@home.nl                                              

// Use the code any way you wish; you assume all responsibility for its 
// correctness and suitability should you use it. 
 
//-------------------------------------------------------------------------------------------------

class cHuffmanNode
{
  friend class cHuffmanTree;

  private:
    U8                       ch;
    U8                       reserved[3];
    U32                      frequency; 
    class cHuffmanNode      *parent, 
                            *left_child, 
                            *right_child;

  public:

   cHuffmanNode ()
   {
     ch=0; 
     frequency=0;
     reserved[0]=reserved[1]=reserved[2]=0;
     parent=left_child=right_child=NULL;
   }

   int                      operator () (const cHuffmanNode& a, const cHuffmanNode& b) const;
   int                      operator () (const cHuffmanNode* a, const cHuffmanNode* b) const;
};

//-------------------------------------------------------------------------------------------------

class cHuffmanTree
{
  // Note that it's up to the client libary to store the tree
  // to use it for later uncompression. Also note that using 
  // the same frequency-table, the tree can be rebuilt, so you
  // can either store the tree, or the frequency-table!
  private:
    cHuffmanNode*              root;
    std::vector<pair<U8,U32> > ft;

  public:

                               cHuffmanTree      () : root (NULL) {};
                               cHuffmanTree      (U32 frequency_table[256]);
                               cHuffmanTree      (std::vector<pair<U8, U32> > frequency_table); 
                               cHuffmanTree      (U8* data, U32 size); 
                              ~cHuffmanTree      ();

    void                       getStringTable    (std::vector<std::string>& bitString);

    U32                        getCompressedSize (U8* udata, U32 len);
    U32                        compress          (U8* udata, U32 len, U8** cdata);
    U32                        uncompress        (U8* cdata, U32 len, U8** udata);
    U32                        compress          (const I8* inname, const I8* outfile);
    U32                        uncompress        (const I8* inname, const I8* outfile);
    U32                        getTree           (std::string& theTree);
    void                       setTree           (const std::string& theTree);

  private:
    bool                       buildTreeDo       (const std::vector<pair<U8, U32> > frequency_table);
    void                       freeTree          (void);
    void                       freeTreeR         (cHuffmanNode* root);
    char*                      getCharacterPath  (cHuffmanNode* n, int c);
    bool                       gcp               (char* s, cHuffmanNode* n, int c);

    bool                       write_bit         (bool bit, U8& theByte, bool reset=false);
    bool                       read_bit          (bool& bit, U8 byte, bool reset=false);

    U32                        getTreeR          (std::string& theTree, cHuffmanNode* node);
    cHuffmanNode*              setTreeR          (const std::string& theTree, U32& idx);

};


#endif

